{ Unit: BcColorButton
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

     Color selection button with colordialog. Custom colors can be
     saved/loaded to/from registry automatically. Use CustomColorGroupName
     to group different buttons to use same CustomColors. If no group is
     given component name is used as group name.


  History:

    2001-11-03: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcColorButton;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, StdCtrls, Buttons,
  Dialogs;

type
  TVAlignment = (vaTop, vaBottom, vaCenter);
  TShowValue = (svFormat, svHTML, svDecimal, svHex, svNone);
  TRegRootKey = (rkHKeyCurrentUser, rkHKeyLocalMachine);

const
  DefaultShowValue = svDecimal;
  DefaultBorderSpace = 2;
  DefaultColorDialog = True;
  DefaultShowValueFormat = '%d, %d, %d';
  DefaultColor = clBtnFace;
  DefaultColorAdjust = True;
  DefaultSaveCustomColors = True;
  DefaultCustomColorsRegKey = '\Software\Bluecave Software\TBcColorButton\';
  DefaultCustomColorsRootKey = rkHKeyCurrentUser;
  DefaultColorDialogOptions{: TColorDialogOptions} = [cdAnyColor];

type
  TBcColorButton = class(TButton)
  private
    { Private declarations }
    FCanvas: TCanvas;
    FColor: TColor;
    FColorAdjust: Boolean;
    FShowValue: TShowValue;
    FBorderSpace: Integer;
    FShowValueFormat: string;
    FColorDialog: Boolean;
    FCustomColorsRegKey: string;
    FSaveCustomColors: Boolean;
    FCustomColorsRootKey: TRegRootKey;
    FCustomColorsGroupName: string;
    FColorDialogOptions: TColorDialogOptions;
    FOnChange: TNotifyEvent;
    procedure SetColor(const Value: TColor);
    procedure WMMeasureItem(var Message: TWMMeasureItem); message CN_MEASUREITEM;
    procedure WMDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure DrawItem(const DrawItemStruct: TDrawItemStruct);
    procedure SetColorAdjust(const Value: Boolean);
    procedure SetShowValue(const Value: TShowValue);
    procedure SetBorderSpace(const Value: Integer);
    procedure SetShowValueFormat(const Value: string);
  protected
    { Protected declarations }
    procedure CreateParams(var Params: TCreateParams); override;
    property Canvas: TCanvas read FCanvas;
    procedure SetButtonStyle(ADefault: Boolean); override;
    procedure LoadCustomColorsRegistry(AColorDialog: TColorDialog);
    procedure SaveCustomColorsRegistry(AColorDialog: TColorDialog);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
  published
    { Published declarations }
    property Color: TColor read FColor write SetColor default DefaultColor;
    property ColorAdjust: Boolean read FColorAdjust write SetColorAdjust default DefaultColorAdjust;
    property ShowValue: TShowValue read FShowValue write SetShowValue default DefaultShowValue;
    property BorderSpace: Integer read FBorderSpace write SetBorderSpace default DefaultBorderSpace;
    property ShowValueFormat: string read FShowValueFormat write SetShowValueFormat;
    property ColorDialog: Boolean read FColorDialog write FColorDialog default DefaultColorDialog;
    property ColorDialogOptions: TColorDialogOptions read FColorDialogOptions write FColorDialogOptions default DefaultColorDialogOptions;
    property SaveCustomColors: Boolean read FSaveCustomColors write FSaveCustomColors default DefaultSaveCustomColors;
    property CustomColorsGroupName: string read FCustomColorsGroupName write FCustomColorsGroupName;
    property CustomColorsRegKey: string read FCustomColorsRegKey write FCustomColorsRegKey;
    property CustomColorsRootKey: TRegRootKey read FCustomColorsRootKey write FCustomColorsRootKey;

    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

uses Registry;

const
  RootKeys: array[TRegRootKey] of Cardinal = (HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE);

{ TBcColorButton }

constructor TBcColorButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TCanvas.Create;
  FColor := DefaultColor;
  FColorAdjust := DefaultColorAdjust;
  FShowValue := DefaultShowValue;
  FBorderSpace := DefaultBorderSpace;
  FShowValueFormat := DefaultShowValueFormat;
  FColorDialog := DefaultColorDialog;
  FColorDialogOptions := DefaultColorDialogOptions;
  FSaveCustomColors := DefaultSaveCustomColors;
  FCustomColorsRegKey := DefaultCustomColorsRegKey;
  FCustomColorsRootKey := DefaultCustomColorsRootKey;

  ControlStyle := ControlStyle + [csReflector];
end;

procedure TBcColorButton.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do Style := Style or BS_OWNERDRAW;
end;

destructor TBcColorButton.Destroy;
begin
  FreeAndNil(FCanvas);
  inherited Destroy;
end;

procedure TBcColorButton.SetColor(const Value: TColor);
begin
  FColor := Value;
  Invalidate;
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TBcColorButton.WMDrawItem(var Message: TWMDrawItem);
begin
  DrawItem(Message.DrawItemStruct^);
end;

procedure TBcColorButton.WMMeasureItem(var Message: TWMMeasureItem);
begin
  with Message.MeasureItemStruct^ do
  begin
    itemWidth := Width;
    itemHeight := Height;
  end;
end;

procedure TBcColorButton.DrawItem(const DrawItemStruct: TDrawItemStruct);
type
  TColorQuad = record
    Red, Green, Blue, Alpha: Byte;
  end;
var
  IsDown, IsDefault, IsDisabled: Boolean;
  R: TRect;
  C: TColor;
  S: string;
  Flags: Longint;
begin
  FCanvas.Handle := DrawItemStruct.hDC;
  R := ClientRect;

  with DrawItemStruct do
  begin
    IsDown := itemState and ODS_SELECTED <> 0;
    IsDefault := itemState and ODS_FOCUS <> 0;
    IsDisabled := DrawItemStruct.itemState and ODS_DISABLED <> 0;
  end;

  Flags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
  if IsDown then Flags := Flags or DFCS_PUSHED;
  if IsDisabled then
    Flags := Flags or DFCS_INACTIVE;

  { DrawFrameControl doesn't allow for drawing a button as the
      default button, so it must be done here. }
  if Focused or IsDefault then
  begin
    FCanvas.Pen.Color := clWindowFrame;
    FCanvas.Pen.Width := 1;
    FCanvas.Brush.Style := bsClear;
    FCanvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
    
    { DrawFrameControl must draw within this border }
    InflateRect(R, -1, -1);
  end;

  if IsDown then
  begin
    FCanvas.Pen.Color := clBtnShadow;
    FCanvas.Pen.Width := 1;
    FCanvas.Brush.Color := clBtnFace;
    FCanvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
    InflateRect(R, -2, -2);
  end else
    DrawFrameControl(DrawItemStruct.hDC, R, DFC_BUTTON, Flags);

{  if Focused and IsDefault then
  begin
    Canvas.Brush.Color := clWindowFrame;
    Canvas.FrameRect(R);
  end;}

  FCanvas.Font := Self.Font;
  C := ColorToRGB(FColor);
  if FColorAdjust then
  begin
    if ((TColorQuad(C).Red + TColorQuad(C).Green + TColorQuad(C).Blue) div 3) < $A0 then
      FCanvas.Font.Color := clWhite
    else
      FCanvas.Font.Color := clBlack;
  end;

  R := ClientRect;
  InflateRect(R, -FBorderSpace - 2, -FBorderSpace - 2);
  if IsDown then OffsetRect(R, 1, 1);
{  if Focused and IsDefault then
  begin}
   if FColorAdjust then
     FCanvas.Pen.Color := FCanvas.Font.Color
   else
     FCanvas.Pen.Color := clWindowFrame;
{  end else
    FCanvas.Pen.Color := FColor;}

  if IsDisabled then
    FCanvas.Brush.Bitmap := AllocPatternBitmap(FColor, clBtnFace)
  else
    FCanvas.Brush.Color := FColor;
  FCanvas.Rectangle(R);
  if IsDisabled then
  begin
    FCanvas.Brush.Bitmap := nil;
  end;

  FCanvas.Brush.Style := bsClear;
  case FShowValue of
    svFormat: S := Format(FShowValueFormat, [TColorQuad(C).Red, TColorQuad(C).Green, TColorQuad(C).Blue]);
    svHTML: S := Format('#%2.2x%2.2x%2.2x', [TColorQuad(C).Red, TColorQuad(C).Green, TColorQuad(C).Blue]);
    svDecimal: S := Format('R: %d, G: %d, B: %d', [TColorQuad(C).Red, TColorQuad(C).Green, TColorQuad(C).Blue]);
    svHex: S := Format('R: %2.2x, G: %2.2x, B: %2.2x', [TColorQuad(C).Red, TColorQuad(C).Green, TColorQuad(C).Blue]);
    svNone: S := Caption;
  end;
  DrawText(FCanvas.Handle, PChar(S), Length(S), R, DT_CENTER or DT_VCENTER or DT_SINGLELINE);

{  if Focused and IsDefault then
  begin
    R := ClientRect;
    InflateRect(R, -4, -4);
    FCanvas.Pen.Color := clWindowFrame;
    FCanvas.Brush.Color := clBtnFace;
    DrawFocusRect(FCanvas.Handle, R);
  end;}

  FCanvas.Handle := 0;
end;

procedure TBcColorButton.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TBcColorButton.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TBcColorButton.SetButtonStyle(ADefault: Boolean);
begin
  Refresh;
end;

procedure TBcColorButton.SetColorAdjust(const Value: Boolean);
begin
  if FColorAdjust <> Value then
  begin
    FColorAdjust := Value;
    Invalidate;
  end;
end;

procedure TBcColorButton.SetShowValue(const Value: TShowValue);
begin
  if FShowValue <> Value then
  begin
    FShowValue := Value;
    Invalidate;
  end;
end;

procedure TBcColorButton.SetBorderSpace(const Value: Integer);
begin
  if FBorderSpace <> Value then
  begin
    FBorderSpace := Value;
    Invalidate;
  end;
end;

procedure TBcColorButton.SetShowValueFormat(const Value: string);
begin
  if FShowValueFormat <> Value then
  begin
    FShowValueFormat := Value;
    if FShowValue = svFormat then
      Invalidate;
  end;
end;

procedure TBcColorButton.Click;
var CD: TColorDialog;
begin
  if FColorDialog then
  begin
    try
      CD := TColorDialog.Create(Self);
      if SaveCustomColors then LoadCustomColorsRegistry(CD);
      CD.Color := FColor;
      CD.Options := FColorDialogOptions;
      if CD.Execute then
      begin
        Color := CD.Color;
        if SaveCustomColors then SaveCustomColorsRegistry(CD);
      end;
    finally
      FreeAndNil(CD);
    end;
  end;
  inherited Click;
end;

procedure TBcColorButton.LoadCustomColorsRegistry(AColorDialog: TColorDialog);
var
  Regs: TRegIniFile;
  C: Char;
  SName: string;
begin
  try
    if FCustomColorsGroupName <> '' then
      SName := FCustomColorsGroupName
    else
      SName := Name;
    Regs := TRegIniFile.Create;
    Regs.RootKey := RootKeys[FCustomColorsRootKey];
    if Regs.OpenKey(ExcludeTrailingBackslash(FCustomColorsRegKey) + '\' + SName, False) then
      for C := 'A' to 'P' do
        AColorDialog.CustomColors.Add(
          Format('Color' + C + '=%s', [Regs.ReadString('', 'Color' + C, '00000000')]));
  finally
    Regs.CloseKey;
    FreeAndNil(Regs);
  end;
end;

procedure TBcColorButton.SaveCustomColorsRegistry(AColorDialog: TColorDialog);
var
  Regs: TRegIniFile;
  I: Integer;
  SName: string;
begin
  try
    if FCustomColorsGroupName <> '' then
      SName := FCustomColorsGroupName
    else
      SName := Name;
    Regs := TRegIniFile.Create;
    Regs.RootKey := RootKeys[FCustomColorsRootKey];
    if Regs.OpenKey(ExcludeTrailingBackslash(FCustomColorsRegKey) + '\' + SName, True) then
      for I := 0 to AColorDialog.CustomColors.Count - 1 do
        Regs.WriteString('',
          AColorDialog.CustomColors.Names[I],
          AColorDialog.CustomColors.Values[AColorDialog.CustomColors.Names[I]]);
  finally
    Regs.CloseKey;
    FreeAndNil(Regs);
  end;
end;

end.
