{ Unit: BcIndyWebUpdater
  ===========================================================================

  The contents of this file are subject to the Bluecave Public License V 1.1
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Auto WebUpdated component. Update your application on the fly. Note, this
    component cannot shutdown/restart your application automatically so you
    should use installer exe to do that. Locked files are updated by using
    RunOnce registry, so it requires reboot. Also if file is set to be removed
    and it's at "Run" line the removing is done with RunOnce key.

    The update file is a plain ini file. It can have Unix linebreak formatting,
    it will be converted before processing it.

  ===========================================================================

    Sections and values available:

    [Updater]
    Name=My App
    Version=1,0,0,0
    QueryMsg=You cannot downgrade after this update.  Continue?
    UpdateURL=http://www.foo.com/NewUpdaterFile.Update
    ShowURL=
    Run=File2, File1

    [Files]
    Count=2

    [File1]
    Version=1,0,0,0
    Location=http://www.foo.com/download/myappinstaller.exe
    Target=<TMP>\myappinstaller.exe
    Description=Main Update
    Remove=1

    [File2]
    Version=1,0,0,0
    Location=http://www.foo.com/download/readme.txt
    Target=<TMP>\readme.txt
    Description=Post Update Readme
    Remove=1


  ===========================================================================

  Description:

    Currently this implementation is simple and does not allow any
    "special" things. Do those through your installer exe.

    All files are downloaded and Run is used to execute files. UpdateURL is
    used as redirector if no [Files] section exists otherwise it is used
    next time update is requested. File removal is done through RunOnce
    registry.

    File versions are compare to the application version. This way you can
    have e.g. file1 with earlier version number than file2. Updater will
    skip file if version is same or even recent. If file does not have version
    number it will use the global version.

    Description is shown in the progress dialog.

    "Run" will run the file using ShellExecute (file associations).

  History:

    2001-12-08: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcIndyWebUpdater;

interface

uses
  Windows, SysUtils, Controls, Classes, BcIndyHTTPFile;

resourcestring
  AppTag = '<APP>';
  TmpTag = '<TMP>';
  WinTag = '<WIN>';
  SysTag = '<SYS>';

const
  DefaultDateFormat = 'yyyy-mm-dd';
  DefaultRegistryKey = 'Software\TBcIndyWebUpdater';

  RegistryKeySuffix = '\WebUpdater';
  RegistryDateName = 'Date';
  RegistryDateFormatName = 'DateFormat';
  RegistryUpdateURLName = 'UpdateURL';

type
  TBcFileVersion = record
    Major, Minor, Release, Build: Word;
  end;

  TBcIndyWebUpdaterMsgStrings = class(TPersistent)
  public
    constructor Create; virtual;
    destructor Destroy; override;
  published
  end;

  TBcIndyWebUpdater = class(TComponent)
  private
    FHTTPFile: TBcIndyHTTPFile;
    FMsgStrings: TBcIndyWebUpdaterMsgStrings;
    FUpdateURL: string;
    FAppRootPath: string;
    FRegistryKey: string;
    FDateFormat: string;
    FDate: TDate;
    FAppFileName: string;
    procedure SetHTTPFile(const Value: TBcIndyHTTPFile);
    procedure SetMsgStrings(const Value: TBcIndyWebUpdaterMsgStrings);
    function GetAppVersion: string;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure GetFile(URL, TargetFile: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetFileVersion: TBcFileVersion;
    function GetTempPath: string;
    function GetTempFileName: string;
    function GetWinPath: string;
    function GetWinSysPath: string;
    function PosEx(Substr: string; S: string; Index: Integer; CaseSensitive: Boolean): Integer;
    function TranslatePath(APath: string): string;
    function ExtractFileNameFromURL(URL: string): string;
    procedure ReadUpdateRegistry;
    procedure WriteUpdateRegistry;

    function RunUpdater: Boolean;

    property AppVersion: string read GetAppVersion;
    property Date: TDate read FDate write FDate;
  published
    property UpdateURL: string read FUpdateURL write FUpdateURL;
    property RegistryKey: string read FRegistryKey write FRegistryKey;
    property HTTPFile: TBcIndyHTTPFile read FHTTPFile write SetHTTPFile;

    property DateFormat: string read FDateFormat write FDateFormat;
    property AppFileName: string read FAppFileName write FAppFileName;

    property AppRootPath: string read FAppRootPath write FAppRootPath;
    property MsgStrings: TBcIndyWebUpdaterMsgStrings read FMsgStrings write SetMsgStrings;
  end;

implementation

uses ShellAPI, Registry, IniFiles;

{ TBcIndyWebUpdater }

constructor TBcIndyWebUpdater.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMsgStrings := TBcIndyWebUpdaterMsgStrings.Create;
  FRegistryKey := DefaultRegistryKey;
  FDateFormat := DefaultDateFormat;
end;

destructor TBcIndyWebUpdater.Destroy;
begin
  FreeAndNil(FMsgStrings);
  inherited Destroy;
end;

function TBcIndyWebUpdater.ExtractFileNameFromURL(URL: string): string;
var I: Integer;
begin
  Result := '';
  for I := Length(URL) downto 1 do
    if URL[I] = '/' then
      Result := Copy(URL, I + 1, Length(URL));
end;

procedure TBcIndyWebUpdater.GetFile(URL, TargetFile: string);
var OldURL: string;
begin
  if Assigned(FHTTPFile) then
  begin
    OldURL := FHTTPFile.URL;
    FHTTPFile.URL := URL;
    FHTTPFile.DownloadFile(TargetFile);
    FHTTPFile.URL := OldURL;
  end;
end;

procedure TBcIndyWebUpdater.ReadUpdateRegistry;
var
  Regs: TRegIniFile;
  APath: string;
begin
  try
    Regs := TRegIniFile.Create('');
    APath := ExcludeTrailingBackslash(FRegistryKey) + RegistryKeySuffix;
    if Regs.OpenKey(APath, False) then
    begin
      try
        FDate := Regs.ReadDate(RegistryDateName);
      except
        if FileExists(FAppFileName) then
          FDate := FileDateToDateTime(FileAge(FAppFileName))
        else
          FDate := SysUtils.Date;
      end;
      FDateFormat := Regs.ReadString(APath, RegistryDateFormatName, FDateFormat);
      FUpdateURL := Regs.ReadString(APath, RegistryUpdateURLName, FUpdateURL);
    end;
  finally
    Regs.CloseKey;
    FreeAndNil(Regs);
  end;
end;

function TBcIndyWebUpdater.GetTempFileName: string;
var ATmpPath: string;
begin
  SetLength(Result, MAX_PATH);
  ATmpPath := GetTempPath;
  Windows.GetTempFileName(PChar(ATmpPath), 'IWU', 0, PChar(Result));
  Result := PChar(Result); // truncate PChar string
end;

function TBcIndyWebUpdater.GetTempPath: string;
begin
  SetLength(Result, MAX_PATH);
  SetLength(Result, Windows.GetTempPath(MAX_PATH, PChar(Result)));
end;

function TBcIndyWebUpdater.GetWinPath: string;
begin
  SetLength(Result, MAX_PATH);
  SetLength(Result, Windows.GetWindowsDirectory(PChar(Result), MAX_PATH));
end;

function TBcIndyWebUpdater.GetWinSysPath: string;
begin
  SetLength(Result, MAX_PATH);
  SetLength(Result, Windows.GetSystemDirectory(PChar(Result), MAX_PATH));
end;

procedure TBcIndyWebUpdater.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

function TBcIndyWebUpdater.PosEx(Substr, S: string;
  Index: Integer; CaseSensitive: Boolean): Integer;
var I, L: Integer;
begin
  Result := 0;
  L := Length(SubStr);
  if L = 0 then Exit;
  if Index = 0 then
    Index := 1;
  for I := Index to Length(S) - L + 1 do
  begin
    if (S[Index] = SubStr[1]) then
    begin
      Result := Index;
      if CaseSensitive then
      begin
        if (AnsiCompareStr(Copy(S, Index, L), SubStr) = 0) then
          Break;
      end else
      begin
        if (AnsiCompareText(Copy(S, Index, L), SubStr) = 0) then
          Break;
      end;
    end;
  end;
end;

function TBcIndyWebUpdater.RunUpdater: Boolean;
begin
  Result := Assigned(FHTTPFile);
  if Result then
  begin
    GetFile(GetRealUpdateURL, GetTempFileName);
  end;
end;

procedure TBcIndyWebUpdater.SetHTTPFile(const Value: TBcIndyHTTPFile);
begin
  FHTTPFile := Value;
end;

procedure TBcIndyWebUpdater.SetMsgStrings(const Value: TBcIndyWebUpdaterMsgStrings);
begin
  FMsgStrings := Value;
//  FMsgStrings.Assign(Value);
end;

function TBcIndyWebUpdater.TranslatePath(APath: string): string;
begin
  APath := StringReplace(APath, AppTag, ExcludeTrailingBackslash(FAppRootPath), [rfReplaceAll, rfIgnoreCase]);
  APath := StringReplace(APath, TmpTag, ExcludeTrailingBackslash(GetTempPath), [rfReplaceAll, rfIgnoreCase]);
  APath := StringReplace(APath, WinTag, GetWinPath, [rfReplaceAll, rfIgnoreCase]);
  APath := StringReplace(APath, SysTag, GetWinSysPath, [rfReplaceAll, rfIgnoreCase]);
end;

procedure TBcIndyWebUpdater.WriteUpdateRegistry;
var
  Regs: TRegIniFile;
  APath: string;
begin
  try
    Regs := TRegIniFile.Create('');
    APath := ExcludeTrailingBackslash(FRegistryKey) + RegistryKeySuffix;
    if Regs.OpenKey(APath, True) then
    begin
      Regs.WriteString(APath, RegistryDateName, FDate);
      Regs.WriteString(APath, RegistryDateFormatName, FDateFormat);
      Regs.WriteString(APath, RegistryUpdateURLName, FUpdateURL);
    end;
  finally
    Regs.CloseKey;
    FreeAndNil(Regs);
  end;
end;

function TBcIndyWebUpdater.GetAppVersion: string;
begin
  if FileExists(FAppFileName) then
  begin
  end;
end;

function TBcIndyWebUpdater.GetFileVersion: TBcFileVersion;
begin
  //
end;

{ TBcIndyWebUpdaterMsgStrings }

constructor TBcIndyWebUpdaterMsgStrings.Create;
begin
  inherited Create;

end;

destructor TBcIndyWebUpdaterMsgStrings.Destroy;
begin
  inherited Destroy;
end;

end.
